@@ -1,11 +1,11 @@
 # Copyright (C) 2009, 2010, 2011, 2012, 2013
 # David Aguilar <davvid@gmail.com>
-"""Provides the main() routine and ColaApplicaiton"""
+"""Provides the main() routine and ColaApplication"""
 from __future__ import division, absolute_import, unicode_literals
 
 import argparse
-import glob
 import os
+import shutil
 import signal
 import sys
 
@@ -17,7 +17,29 @@ if sys.platform == 'darwin':
     if os.path.isdir(homebrew_mods):
         sys.path.append(homebrew_mods)
 
-import sip
+
+errmsg = """Sorry, you do not seem to have PyQt4 installed.
+Please install it before using git-cola.
+e.g.: sudo apt-get install python-qt4
+"""
+
+# /usr/include/sysexits.h
+#define EX_OK           0   /* successful termination */
+#define EX_USAGE        64  /* command line usage error */
+#define EX_NOINPUT      66  /* cannot open input */
+#define EX_UNAVAILABLE  69  /* service unavailable */
+EX_OK = 0
+EX_USAGE = 64
+EX_NOINPUT = 66
+EX_UNAVAILABLE = 69
+
+
+try:
+    import sip
+except ImportError:
+    sys.stderr.write(errmsg)
+    sys.exit(EX_UNAVAILABLE)
+
 sip.setapi('QString', 1)
 sip.setapi('QDate', 1)
 sip.setapi('QDateTime', 1)
@@ -27,14 +49,14 @@ sip.setapi('QUrl', 1)
 sip.setapi('QVariant', 1)
 
 try:
-    from PyQt4 import QtGui
     from PyQt4 import QtCore
-    from PyQt4.QtCore import SIGNAL
 except ImportError:
-    sys.stderr.write('Sorry, you do not seem to have PyQt4 installed.\n')
-    sys.stderr.write('Please install it before using git-cola.\n')
-    sys.stderr.write('e.g.: sudo apt-get install python-qt4\n')
-    sys.exit(-1)
+    sys.stderr.write(errmsg)
+    sys.exit(EX_UNAVAILABLE)
+
+from PyQt4 import QtGui
+from PyQt4.QtCore import Qt
+from PyQt4.QtCore import SIGNAL
 
 # Import cola modules
 from cola import cmds
@@ -42,6 +64,7 @@ from cola import core
 from cola import compat
 from cola import git
 from cola import gitcfg
+from cola import icons
 from cola import inotify
 from cola import i18n
 from cola import qtcompat
@@ -64,7 +87,7 @@ def setup_environment():
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
     # Session management wants an absolute path when restarting
-    sys.argv[0] = core.abspath(sys.argv[0])
+    sys.argv[0] = sys_argv0 = core.abspath(sys.argv[0])
 
     # Spoof an X11 display for SSH
     os.environ.setdefault('DISPLAY', ':0')
@@ -77,8 +100,8 @@ def setup_environment():
 
     # Setup the path so that git finds us when we run 'git cola'
     path_entries = core.getenv('PATH', '').split(os.pathsep)
-    bindir = os.path.dirname(core.abspath(__file__))
-    path_entries.insert(0, bindir)
+    bindir = os.path.dirname(sys_argv0)
+    path_entries.append(bindir)
     path = os.pathsep.join(path_entries)
     compat.setenv('PATH', path)
 
@@ -140,37 +163,25 @@ class ColaApplication(object):
     ColaApplication handles i18n of user-visible data
     """
 
-    def __init__(self, argv, locale=None, gui=True, git_path=None):
+    def __init__(self, argv, locale=None, gui=True):
         cfgactions.install()
         i18n.install(locale)
         qtcompat.install()
         qtutils.install()
+        icons.install()
 
         self.notifier = QtCore.QObject()
-        self.notifier.connect(self.notifier, SIGNAL('update_files'), self._update_files)
+        self.notifier.connect(self.notifier, SIGNAL('update_files()'),
+                              self._update_files, Qt.QueuedConnection)
         # Call _update_files when inotify detects changes
         inotify.observer(self._update_files_notifier)
 
-        # Add the default style dir so that we find our icons
-        icon_dir = resources.icon_dir()
-        qtcompat.add_search_path(os.path.basename(icon_dir), icon_dir)
-
         if gui:
-            self._app = current(tuple(argv), git_path)
-            self._app.setWindowIcon(qtutils.git_icon())
+            self._app = current(tuple(argv))
+            self._app.setWindowIcon(icons.cola())
         else:
             self._app = QtCore.QCoreApplication(argv)
 
-        self._app.setStyleSheet("""
-            QMainWindow::separator {
-                width: 3px;
-                height: 3px;
-            }
-            QMainWindow::separator:hover {
-                background: white;
-            }
-            """)
-
     def activeWindow(self):
         """Wrap activeWindow()"""
         return self._app.activeWindow()
@@ -191,19 +202,18 @@ class ColaApplication(object):
         cmds.do(cmds.Refresh)
 
     def _update_files_notifier(self):
-        self.notifier.emit(SIGNAL('update_files'))
+        self.notifier.emit(SIGNAL('update_files()'))
 
 
 @memoize
-def current(argv, git_path=None):
-    return ColaQApplication(list(argv), git_path)
+def current(argv):
+    return ColaQApplication(list(argv))
 
 
 class ColaQApplication(QtGui.QApplication):
 
-    def __init__(self, argv, git_path=None):
+    def __init__(self, argv):
         QtGui.QApplication.__init__(self, argv)
-        self.git_path = git_path
         self.view = None ## injected by application_start()
 
     def event(self, e):
@@ -220,8 +230,7 @@ class ColaQApplication(QtGui.QApplication):
         sid = ustr(session_mgr.sessionId())
         skey = ustr(session_mgr.sessionKey())
         session_id = '%s_%s' % (sid, skey)
-        session = Session(session_id,
-                          repo=core.getcwd(), git_path=self.git_path)
+        session = Session(session_id, repo=core.getcwd())
         self.view.save_state(settings=session)
 
 
@@ -229,17 +238,11 @@ def process_args(args):
     if args.version:
         # Accept 'git cola --version' or 'git cola version'
         version.print_version()
-        sys.exit(0)
+        sys.exit(EX_OK)
 
     # Handle session management
     restore_session(args)
 
-    if args.git_path:
-        # Adds git to the PATH.  This is needed on Windows.
-        path_entries = core.getenv('PATH', '').split(os.pathsep)
-        path_entries.insert(0, os.path.dirname(core.decode(args.git_path)))
-        compat.setenv('PATH', os.pathsep.join(path_entries))
-
     # Bail out if --repo is not a directory
     repo = core.decode(args.repo)
     if repo.startswith('file:'):
@@ -249,7 +252,7 @@ def process_args(args):
         errmsg = N_('fatal: "%s" is not a directory.  '
                     'Please specify a correct --repo <path>.') % repo
         core.stderr(errmsg)
-        sys.exit(-1)
+        sys.exit(EX_USAGE)
 
     # We do everything relative to the repo root
     os.chdir(args.repo)
@@ -265,7 +268,6 @@ def restore_session(args):
     if session.load():
         args.settings = session
         args.repo = session.repo
-        args.git_path = session.git_path
 
 
 def application_init(args, update=False):
@@ -277,7 +279,8 @@ def application_init(args, update=False):
     process_args(args)
 
     app = new_application(args)
-    model = new_model(app, args.repo, prompt=args.prompt)
+    model = new_model(app, args.repo,
+                      prompt=args.prompt, settings=args.settings)
     if update:
         model.update_status()
     cfg = gitcfg.current()
@@ -294,7 +297,8 @@ def application_start(context, view):
     view.raise_()
 
     # Scan for the first time
-    task = _start_update_thread(context.model)
+    runtask = qtutils.RunTask(parent=view)
+    init_update_task(view, runtask, context.model)
 
     # Start the inotify thread
     inotify.start()
@@ -310,11 +314,9 @@ def application_start(context, view):
     # All done, cleanup
     inotify.stop()
     QtCore.QThreadPool.globalInstance().waitForDone()
-    del task
 
-    pattern = utils.tmp_file_pattern()
-    for filename in glob.glob(pattern):
-        os.unlink(filename)
+    tmpdir = utils.tmpdir()
+    shutil.rmtree(tmpdir, ignore_errors=True)
 
     return result
 
@@ -332,10 +334,6 @@ def add_common_arguments(parser):
     parser.add_argument('--prompt', action='store_true', default=False,
                         help='prompt for a repository')
 
-    # Used on Windows for adding 'git' to the path
-    parser.add_argument('-g', '--git-path', metavar='<path>', default=None,
-                        help='use the specified git executable')
-
     # Resume an X Session Management session
     parser.add_argument('-session', metavar='<session>', default=None,
                         help=argparse.SUPPRESS)
@@ -343,17 +341,18 @@ def add_common_arguments(parser):
 
 def new_application(args):
     # Initialize the app
-    return ColaApplication(sys.argv, git_path=args.git_path)
+    return ColaApplication(sys.argv)
 
 
-def new_model(app, repo, prompt=False):
+def new_model(app, repo, prompt=False, settings=None):
     model = main.model()
     valid = model.set_worktree(repo) and not prompt
     while not valid:
-        startup_dlg = startup.StartupDialog(app.activeWindow())
+        startup_dlg = startup.StartupDialog(app.activeWindow(),
+                                            settings=settings)
         gitdir = startup_dlg.find_git_repo()
         if not gitdir:
-            sys.exit(-1)
+            sys.exit(EX_NOINPUT)
         valid = model.set_worktree(gitdir)
 
     # Finally, go to the root of the git repo
@@ -361,29 +360,23 @@ def new_model(app, repo, prompt=False):
     return model
 
 
-def _start_update_thread(model):
+def init_update_task(parent, runtask, model):
     """Update the model in the background
 
     git-cola should startup as quickly as possible.
 
     """
-    class UpdateTask(QtCore.QRunnable):
-        def run(self):
-            model.update_status(update_index=True)
 
-    # Hold onto a reference to prevent PyQt from dereferencing
-    task = UpdateTask()
-    QtCore.QThreadPool.globalInstance().start(task)
+    def update_status():
+        model.update_status(update_index=True)
 
-    return task
+    task = qtutils.SimpleTask(parent, update_status)
+    runtask.start(task)
 
 
 def _send_msg():
     if git.GIT_COLA_TRACE == 'trace':
-        msg = ('info: Trace enabled.  '
-               'Many of commands reported with "trace" use git\'s stable '
-               '"plumbing" API and are not intended for typical '
-               'day-to-day use.  Here be dragons')
+        msg = 'info: debug mode enabled using GIT_COLA_TRACE=trace'
         Interaction.log(msg)
 
 
