@@ -1,6 +1,7 @@
 from __future__ import division, absolute_import, unicode_literals
 
 import os
+import re
 import subprocess
 import sys
 from fnmatch import fnmatch
@@ -15,6 +16,7 @@ from cola import compat
 from cola import core
 from cola import gitcfg
 from cola import gitcmds
+from cola import icons
 from cola import inotify
 from cola import utils
 from cola import difftool
@@ -144,7 +146,6 @@ class Command(ModelCommand):
 class AmendMode(Command):
     """Try to amend a commit."""
 
-    SHORTCUT = 'Ctrl+M'
     LAST_MESSAGE = None
 
     @staticmethod
@@ -357,19 +358,19 @@ class ResetMode(Command):
 class Commit(ResetMode):
     """Attempt to create a new commit."""
 
-    SHORTCUT = 'Ctrl+Return'
-
-    def __init__(self, amend, msg, sign):
+    def __init__(self, amend, msg, sign, no_verify=False):
         ResetMode.__init__(self)
         self.amend = amend
         self.msg = msg
         self.sign = sign
+        self.no_verify = no_verify
         self.old_commitmsg = self.model.commitmsg
         self.new_commitmsg = ''
 
     def do(self):
         # Create the commit message file
-        msg = self.strip_comments(self.msg)
+        comment_char = prefs.comment_char()
+        msg = self.strip_comments(self.msg, comment_char=comment_char)
         tmpfile = utils.tmp_filename('commit-message')
         try:
             core.write(tmpfile, msg)
@@ -378,7 +379,8 @@ class Commit(ResetMode):
             status, out, err = self.model.git.commit(F=tmpfile,
                                                      v=True,
                                                      gpg_sign=self.sign,
-                                                     amend=self.amend)
+                                                     amend=self.amend,
+                                                     no_verify=self.no_verify)
         finally:
             core.unlink(tmpfile)
 
@@ -393,10 +395,10 @@ class Commit(ResetMode):
         return status, out, err
 
     @staticmethod
-    def strip_comments(msg):
+    def strip_comments(msg, comment_char='#'):
         # Strip off comments
         message_lines = [line for line in msg.split('\n')
-                            if not line.startswith('#')]
+                            if not line.startswith(comment_char)]
         msg = '\n'.join(message_lines)
         if not msg.endswith('\n'):
             msg += '\n'
@@ -426,8 +428,8 @@ class Ignore(Command):
 
 def file_summary(files):
     txt = subprocess.list2cmdline(files)
-    if len(txt) > 2048:
-        txt = txt[:2048].rstrip() + '...'
+    if len(txt) > 768:
+        txt = txt[:768].rstrip() + '...'
     return txt
 
 
@@ -572,9 +574,6 @@ class RemoveFiles(Command):
 class Delete(RemoveFiles):
     """Delete files."""
 
-    SHORTCUT = 'Ctrl+Shift+Backspace'
-    ALT_SHORTCUT = 'Ctrl+Backspace'
-
     def __init__(self, filenames):
         RemoveFiles.__init__(self, os.remove, filenames)
 
@@ -590,8 +589,7 @@ class Delete(RemoveFiles):
         ok_txt = N_('Delete Files')
 
         if not Interaction.confirm(title, msg, info_txt, ok_txt,
-                                   default=True,
-                                   icon=resources.icon('remove.svg')):
+                                   default=True, icon=icons.remove()):
             return
 
         return RemoveFiles.do(self)
@@ -600,7 +598,6 @@ class Delete(RemoveFiles):
 class MoveToTrash(RemoveFiles):
     """Move files to the trash using send2trash"""
 
-    SHORTCUT = 'Ctrl+Backspace'
     AVAILABLE = send2trash is not None
 
     def __init__(self, filenames):
@@ -730,7 +727,6 @@ class Difftool(Command):
 
 class Edit(Command):
     """Edit a file using the configured gui.editor."""
-    SHORTCUT = 'Ctrl+E'
 
     @staticmethod
     def name():
@@ -791,8 +787,6 @@ class FormatPatch(Command):
 
 class LaunchDifftool(BaseCommand):
 
-    SHORTCUT = 'Ctrl+D'
-
     @staticmethod
     def name():
         return N_('Launch Diff Tool')
@@ -819,8 +813,6 @@ class LaunchDifftool(BaseCommand):
 
 class LaunchTerminal(BaseCommand):
 
-    SHORTCUT = 'Ctrl+T'
-
     @staticmethod
     def name():
         return N_('Launch Terminal')
@@ -838,7 +830,6 @@ class LaunchTerminal(BaseCommand):
 
 
 class LaunchEditor(Edit):
-    SHORTCUT = 'Ctrl+E'
 
     @staticmethod
     def name():
@@ -933,9 +924,8 @@ class Merge(Command):
         no_ff = self.no_ff
         no_commit = self.no_commit
         sign = self.sign
-        msg = gitcmds.merge_message(revision)
 
-        status, out, err = self.model.git.merge('-m', msg, revision,
+        status, out, err = self.model.git.merge(revision,
                                                 gpg_sign=sign,
                                                 no_ff=no_ff,
                                                 no_commit=no_commit,
@@ -946,7 +936,6 @@ class Merge(Command):
 
 class OpenDefaultApp(BaseCommand):
     """Open a file using the OS default."""
-    SHORTCUT = 'Space'
 
     @staticmethod
     def name():
@@ -969,7 +958,6 @@ class OpenDefaultApp(BaseCommand):
 
 class OpenParentDir(OpenDefaultApp):
     """Open parent directories using the OS default."""
-    SHORTCUT = 'Shift+Space'
 
     @staticmethod
     def name():
@@ -1048,50 +1036,90 @@ class Clone(Command):
         return self
 
 
+def unix_path(path, is_win32=utils.is_win32):
+    """Git for Windows requires unix paths, so force them here
+    """
+    unix_path = path
+    if is_win32():
+        first = path[0]
+        second = path[1]
+        if second == ':': # sanity check, this better be a Windows-style path
+            unix_path = '/' + first + path[2:].replace('\\', '/')
+
+    return unix_path
+
+
 class GitXBaseContext(object):
 
     def __init__(self, **kwargs):
-        self.extras = kwargs
+        self.env = {'GIT_EDITOR': prefs.editor()}
+        self.env.update(kwargs)
 
     def __enter__(self):
         compat.setenv('GIT_SEQUENCE_EDITOR',
-                      resources.share('bin', 'git-xbase'))
-        for var, value in self.extras.items():
+                      unix_path(resources.share('bin', 'git-xbase')))
+        for var, value in self.env.items():
             compat.setenv(var, value)
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         compat.unsetenv('GIT_SEQUENCE_EDITOR')
-        for var in self.extras:
+        for var in self.env:
             compat.unsetenv(var)
 
 
 class Rebase(Command):
 
-    def __init__(self, branch, capture_output=True):
+    def __init__(self,
+                 upstream=None, branch=None, capture_output=True, **kwargs):
+        """Start an interactive rebase session
+
+        :param upstream: upstream branch
+        :param branch: optional branch to checkout
+        :param capture_output: whether to capture stdout and stderr
+        :param kwargs: forwarded directly to `git.rebase()`
+
+        """
         Command.__init__(self)
+
+        self.upstream = upstream
         self.branch = branch
         self.capture_output = capture_output
+        self.kwargs = kwargs
+
+    def prepare_arguments(self):
+        args = []
+        kwargs = {}
+
+        if self.capture_output:
+            kwargs['_stderr'] = None
+            kwargs['_stdout'] = None
+
+        # Rebase actions must be the only option specified
+        for action in ('continue', 'abort', 'skip', 'edit_todo'):
+            if self.kwargs.get(action, False):
+                kwargs[action] = self.kwargs[action]
+                return args, kwargs
+
+        kwargs['interactive'] = True
+        kwargs['autosquash'] = self.kwargs.get('autosquash', True)
+        kwargs.update(self.kwargs)
+
+        if self.upstream:
+            args.append(self.upstream)
+        if self.branch:
+            args.append(self.branch)
+
+        return args, kwargs
 
     def do(self):
-        status = 1
-        out = ''
-        err = ''
-        extra = {}
-        branch = self.branch
-        if not branch:
-            return status, out, err
-        if self.capture_output:
-            extra['_stderr'] = None
-            extra['_stdout'] = None
+        (status, out, err) = (1, '', '')
+        args, kwargs = self.prepare_arguments()
+        upstream_title = self.upstream or '@{upstream}'
         with GitXBaseContext(
-                GIT_EDITOR=prefs.editor(),
-                GIT_XBASE_TITLE=N_('Rebase onto %s') % branch,
+                GIT_XBASE_TITLE=N_('Rebase onto %s') % upstream_title,
                 GIT_XBASE_ACTION=N_('Rebase')):
-            status, out, err = self.model.git.rebase(branch,
-                                                     interactive=True,
-                                                     autosquash=True,
-                                                     **extra)
+            status, out, err = self.model.git.rebase(*args, **kwargs)
         Interaction.log_status(status, out, err)
         self.model.update_status()
         return status, out, err
@@ -1100,28 +1128,40 @@ class Rebase(Command):
 class RebaseEditTodo(Command):
 
     def do(self):
+        (status, out, err) = (1, '', '')
         with GitXBaseContext(
                 GIT_XBASE_TITLE=N_('Edit Rebase'),
                 GIT_XBASE_ACTION=N_('Save')):
             status, out, err = self.model.git.rebase(edit_todo=True)
         Interaction.log_status(status, out, err)
         self.model.update_status()
+        return status, out, err
 
 
 class RebaseContinue(Command):
 
     def do(self):
-        status, out, err = self.model.git.rebase('--continue')
+        (status, out, err) = (1, '', '')
+        with GitXBaseContext(
+                GIT_XBASE_TITLE=N_('Rebase'),
+                GIT_XBASE_ACTION=N_('Rebase')):
+            status, out, err = self.model.git.rebase('--continue')
         Interaction.log_status(status, out, err)
         self.model.update_status()
+        return status, out, err
 
 
 class RebaseSkip(Command):
 
     def do(self):
-        status, out, err = self.model.git.rebase(skip=True)
+        (status, out, err) = (1, '', '')
+        with GitXBaseContext(
+                GIT_XBASE_TITLE=N_('Rebase'),
+                GIT_XBASE_ACTION=N_('Rebase')):
+            status, out, err = self.model.git.rebase(skip=True)
         Interaction.log_status(status, out, err)
         self.model.update_status()
+        return status, out, err
 
 
 class RebaseAbort(Command):
@@ -1142,8 +1182,6 @@ class Rescan(Command):
 class Refresh(Command):
     """Update refs and refresh the index"""
 
-    SHORTCUTS = ('Ctrl+R', 'F5')
-
     @staticmethod
     def name():
         return N_('Refresh')
@@ -1157,7 +1195,7 @@ class RevertEditsCommand(ConfirmAction):
     def __init__(self):
         ConfirmAction.__init__(self)
         self.model = main.model()
-        self.icon = resources.icon('undo.svg')
+        self.icon = icons.undo()
 
     def ok_to_run(self):
         return self.model.undoable()
@@ -1191,8 +1229,6 @@ class RevertEditsCommand(ConfirmAction):
 
 class RevertUnstagedEdits(RevertEditsCommand):
 
-    SHORTCUT = 'Ctrl+U'
-
     @staticmethod
     def name():
         return N_('Revert Unstaged Edits...')
@@ -1215,8 +1251,6 @@ class RevertUnstagedEdits(RevertEditsCommand):
 
 class RevertUncommittedEdits(RevertEditsCommand):
 
-    SHORTCUT = 'Ctrl+Z'
-
     @staticmethod
     def name():
         return N_('Revert Uncommitted Edits...')
@@ -1345,7 +1379,6 @@ class ShowUntracked(Command):
 
 
 class SignOff(Command):
-    SHORTCUT = 'Ctrl+I'
 
     @staticmethod
     def name():
@@ -1378,9 +1411,54 @@ class SignOff(Command):
         return '\nSigned-off-by: %s <%s>' % (name, email)
 
 
+def check_conflicts(unmerged):
+    """Check paths for conflicts
+
+    Conflicting files can be filtered out one-by-one.
+
+    """
+    if prefs.check_conflicts():
+        unmerged = [path for path in unmerged if is_conflict_free(path)]
+    return unmerged
+
+
+def is_conflict_free(path):
+    """Return True if `path` contains no conflict markers
+    """
+    rgx = re.compile(r'^(<<<<<<<|\|\|\|\|\|\|\||>>>>>>>) ')
+    try:
+        with core.xopen(path, 'r') as f:
+            for line in f:
+                line = core.decode(line, errors='ignore')
+                if rgx.match(line):
+                    if should_stage_conflicts(path):
+                        return True
+                    else:
+                        return False
+    except IOError:
+        # We can't read this file ~ we may be staging a removal
+        pass
+    return True
+
+
+def should_stage_conflicts(path):
+    """Inform the user that a file contains merge conflicts
+
+    Return `True` if we should stage the path nonetheless.
+
+    """
+    title = msg = N_('Stage conflicts?')
+    info = N_('%s appears to contain merge conflicts.\n\n'
+              'You should probably skip this file.\n'
+              'Stage it anyways?') % path
+    ok_text = N_('Stage conflicts')
+    cancel_text = N_('Skip')
+    return Interaction.confirm(title, msg, info, ok_text,
+                               default=False, cancel_text=cancel_text)
+
+
 class Stage(Command):
     """Stage a set of paths."""
-    SHORTCUT = 'Ctrl+S'
 
     @staticmethod
     def name():
@@ -1401,53 +1479,68 @@ class Stage(Command):
             self.model.stage_paths(self.paths)
 
 
-class StageModified(Stage):
-    """Stage all modified files."""
+class StageCarefully(Stage):
+    """Only stage when the path list is non-empty
 
-    SHORTCUT = 'Ctrl+S'
+    We use "git add -u -- <pathspec>" to stage, and it stages everything by
+    default when no pathspec is specified, so this class ensures that paths
+    are specified before calling git.
+
+    When no paths are specified, the command does nothing.
+
+    """
+    def __init__(self):
+        Stage.__init__(self, None)
+        self.init_paths()
+
+    def init_paths(self):
+        pass
+
+    def ok_to_run(self):
+        """Prevent catch-all "git add -u" from adding unmerged files"""
+        return self.paths or not self.model.unmerged
+
+    def do(self):
+        if self.ok_to_run():
+            Stage.do(self)
+
+
+class StageModified(StageCarefully):
+    """Stage all modified files."""
 
     @staticmethod
     def name():
         return N_('Stage Modified')
 
-    def __init__(self):
-        Stage.__init__(self, None)
+    def init_paths(self):
         self.paths = self.model.modified
 
 
-class StageUnmerged(Stage):
-    """Stage all modified files."""
-
-    SHORTCUT = 'Ctrl+S'
+class StageUnmerged(StageCarefully):
+    """Stage unmerged files."""
 
     @staticmethod
     def name():
         return N_('Stage Unmerged')
 
-    def __init__(self):
-        Stage.__init__(self, None)
-        self.paths = self.model.unmerged
+    def init_paths(self):
+        self.paths = check_conflicts(self.model.unmerged)
 
 
-class StageUntracked(Stage):
+class StageUntracked(StageCarefully):
     """Stage all untracked files."""
 
-    SHORTCUT = 'Ctrl+S'
-
     @staticmethod
     def name():
         return N_('Stage Untracked')
 
-    def __init__(self):
-        Stage.__init__(self, None)
+    def init_paths(self):
         self.paths = self.model.untracked
 
 
 class StageOrUnstage(Command):
     """If the selection is staged, unstage it, otherwise stage"""
 
-    SHORTCUT = 'Ctrl+S'
-
     @staticmethod
     def name():
         return N_('Stage / Unstage')
@@ -1458,8 +1551,9 @@ class StageOrUnstage(Command):
             do(Unstage, s.staged)
 
         unstaged = []
-        if s.unmerged:
-            unstaged.extend(s.unmerged)
+        unmerged = check_conflicts(s.unmerged)
+        if unmerged:
+            unstaged.extend(unmerged)
         if s.modified:
             unstaged.extend(s.modified)
         if s.untracked:
@@ -1510,8 +1604,6 @@ class Tag(Command):
 class Unstage(Command):
     """Unstage a set of paths."""
 
-    SHORTCUT = 'Ctrl+S'
-
     @staticmethod
     def name():
         return N_('Unstage')
