@@ -4,12 +4,13 @@ import copy
 import fnmatch
 import os
 import re
+import struct
+from binascii import unhexlify
 from os.path import join
 
 from cola import core
 from cola import git
 from cola import observable
-from cola import utils
 from cola.decorators import memoize
 from cola.git import STDOUT
 from cola.compat import ustr
@@ -375,9 +376,20 @@ class GitConfig(observable.Observable):
         if not term:
             # find a suitable default terminal
             term = 'xterm -e' # for mac osx
-            candidates = ('gnome-terminal', 'xfce4-terminal', 'konsole')
+            candidates = ('xfce4-terminal', 'konsole')
             for basename in candidates:
                 if core.exists('/usr/bin/%s' % basename):
                     term = '%s -e' % basename
                     break
         return term
+
+    def color(self, key, default):
+        string = self.get('cola.color.%s' % key, default)
+        string = core.encode(string)
+        default = core.encode(default)
+        struct_layout = core.encode('BBB')
+        try:
+            r, g, b = struct.unpack(struct_layout, unhexlify(string))
+        except Exception:
+            r, g, b = struct.unpack(struct_layout, unhexlify(default))
+        return (r, g, b)
