@@ -13,6 +13,9 @@ from cola.git import STDOUT
 from cola.i18n import N_
 
 
+EMPTY_TREE_SHA1 = '4b825dc642cb6eb9a060e54bf8d69288fbee4904'
+
+
 class InvalidRepositoryError(Exception):
     pass
 
@@ -54,18 +57,19 @@ def _parse_diff_filenames(out):
         return []
 
 
-def tracked_files():
+def tracked_files(*args):
     """Return the names of all files in the repository"""
-    out = git.ls_files(z=True)[STDOUT]
+    out = git.ls_files('--', *args, z=True)[STDOUT]
     if out:
         return sorted(out[:-1].split('\0'))
     else:
         return []
 
 
-def all_files():
+def all_files(*args):
     """Returns a sorted list of all files, including untracked files."""
-    ls_files = git.ls_files(z=True,
+    ls_files = git.ls_files('--', *args,
+                            z=True,
                             cached=True,
                             others=True,
                             exclude_standard=True)[STDOUT]
@@ -90,7 +94,9 @@ def current_branch():
         if _current_branch.key == key:
             return _current_branch.value
     except OSError:
-        pass
+        # OSError means we can't use the stat cache
+        key = 0
+
     status, data, err = git.rev_parse('HEAD', symbolic_full_name=True)
     if status != 0:
         # git init -- read .git/HEAD.  We could do this unconditionally...
@@ -194,9 +200,9 @@ def untracked_files(git=git, paths=None):
 
     if paths is None:
         paths = []
-    filter_paths = ['--'] + paths
+    args = ['--'] + paths
     out = git.ls_files(z=True, others=True, exclude_standard=True,
-                       *filter_paths)[STDOUT]
+                       *args)[STDOUT]
     if out:
         return out[:-1].split('\0')
     return []
@@ -208,7 +214,8 @@ def tag_list():
 
 
 def log(git, *args, **kwargs):
-    return git.log(no_color=True, no_ext_diff=True, *args, **kwargs)[STDOUT]
+    return git.log(no_color=True, no_abbrev_commit=True,
+                   no_ext_diff=True, *args, **kwargs)[STDOUT]
 
 
 def commit_diff(sha1, git=git):
@@ -276,6 +283,7 @@ def diff_helper(commit=None,
                 endref=None,
                 filename=None,
                 cached=True,
+                deleted=False,
                 head=None,
                 amending=False,
                 with_diff_header=False,
@@ -304,11 +312,6 @@ def diff_helper(commit=None,
             cfg = gitcfg.current()
             encoding = cfg.file_encoding(filename)
 
-    if filename is not None:
-        deleted = cached and not core.exists(filename)
-    else:
-        deleted = False
-
     status, out, err = git.diff(R=reverse, M=True, cached=cached,
                                 _encoding=encoding,
                                 *argv,
@@ -429,25 +432,10 @@ def untrack_paths(args, head='HEAD'):
     return git.update_index('--', force_remove=True, *set(args))
 
 
-def worktree_state(head='HEAD'):
-    """Return a tuple of files in various states of being
-
-    Can be staged, unstaged, untracked, unmerged, or changed
-    upstream.
-
-    """
-    state = worktree_state_dict(head=head)
-    return(state.get('staged', []),
-           state.get('modified', []),
-           state.get('unmerged', []),
-           state.get('untracked', []),
-           state.get('upstream_changed', []))
-
-
-def worktree_state_dict(head='HEAD',
-                        update_index=False,
-                        display_untracked=True,
-                        paths=None):
+def worktree_state(head='HEAD',
+                   update_index=False,
+                   display_untracked=True,
+                   paths=None):
     """Return a dict of files in various states of being
 
     :rtype: dict, keys are staged, unstaged, untracked, unmerged,
@@ -457,23 +445,15 @@ def worktree_state_dict(head='HEAD',
     if update_index:
         git.update_index(refresh=True)
 
-    staged, unmerged, staged_submods = diff_index(head, paths=paths)
-    modified, modified_submods = diff_worktree(paths)
+    staged, unmerged, staged_deleted, staged_submods = diff_index(head,
+                                                                  paths=paths)
+    modified, unstaged_deleted, modified_submods = diff_worktree(paths)
     untracked = display_untracked and untracked_files(paths=paths) or []
 
     # Remove unmerged paths from the modified list
-    unmerged_set = set(unmerged)
-    modified_set = set(modified)
-    modified_unmerged = modified_set.intersection(unmerged_set)
-    for path in modified_unmerged:
-        modified.remove(path)
-
-    # All submodules
-    submodules = staged_submods.union(modified_submods)
-
-    # Add submodules to the staged and unstaged lists
-    staged.extend(staged_submods)
-    modified.extend(modified_submods)
+    if unmerged:
+        unmerged_set = set(unmerged)
+        modified = [path for path in modified if path not in unmerged_set]
 
     # Look for upstream modified files if this is a tracking branch
     upstream_changed = diff_upstream(head)
@@ -490,62 +470,65 @@ def worktree_state_dict(head='HEAD',
             'unmerged': unmerged,
             'untracked': untracked,
             'upstream_changed': upstream_changed,
-            'submodules': submodules}
+            'staged_deleted': staged_deleted,
+            'unstaged_deleted': unstaged_deleted,
+            'submodules': staged_submods | modified_submods}
+
+
+def _parse_raw_diff(out):
+    while out:
+        info, path, out = out.split('\0', 2)
+        status = info[-1]
+        is_submodule = ('160000' in info[1:14])
+        yield (path, status, is_submodule)
 
 
 def diff_index(head, cached=True, paths=None):
-    submodules = set()
     staged = []
     unmerged = []
+    deleted = set()
+    submodules = set()
 
     if paths is None:
         paths = []
-    filter_paths = [head, '--'] + paths
-    status, out, err = git.diff_index(cached=cached, z=True,
-                                      *filter_paths)
+    args = [head, '--'] + paths
+    status, out, err = git.diff_index(cached=cached, z=True, *args)
     if status != 0:
         # handle git init
-        return tracked_files(), unmerged, submodules
+        args[0] = EMPTY_TREE_SHA1
+        status, out, err = git.diff_index(cached=cached, z=True, *args)
 
-    while out:
-        rest, out = out.split('\0', 1)
-        name, out = out.split('\0', 1)
-        status = rest[-1]
-        if '160000' in rest[1:14]:
-            submodules.add(name)
-        elif status  in 'DAMT':
-            staged.append(name)
+    for path, status, is_submodule in _parse_raw_diff(out):
+        if is_submodule:
+            submodules.add(path)
+        if status in 'DAMT':
+            staged.append(path)
+            if status == 'D':
+                deleted.add(path)
         elif status == 'U':
-            unmerged.append(name)
+            unmerged.append(path)
 
-    return staged, unmerged, submodules
+    return staged, unmerged, deleted, submodules
 
 
 def diff_worktree(paths=None):
     modified = []
+    deleted = set()
     submodules = set()
 
     if paths is None:
         paths = []
-    filter_paths = ['--'] + paths
-    status, out, err = git.diff_files(z=True, *filter_paths)
-    if status != 0:
-        # handle git init
-        out = git.ls_files(modified=True, z=True)[STDOUT]
-        if out:
-            modified = out[:-1].split('\0')
-        return modified, submodules
+    args = ['--'] + paths
+    status, out, err = git.diff_files(z=True, *args)
+    for path, status, is_submodule in _parse_raw_diff(out):
+        if is_submodule:
+            submodules.add(path)
+        if status in 'DAMT':
+            modified.append(path)
+            if status == 'D':
+                deleted.add(path)
 
-    while out:
-        rest, out = out.split('\0', 1)
-        name, out = out.split('\0', 1)
-        status = rest[-1]
-        if '160000' in rest[1:14]:
-            submodules.add(name)
-        elif status in 'DAMT':
-            modified.append(name)
-
-    return modified, submodules
+    return modified, deleted, submodules
 
 
 def diff_upstream(head):
@@ -666,9 +649,9 @@ def abort_merge():
         merge_msg_path = merge_message_path()
 
 
-def merge_message(revision):
-    """Return a merge message for FETCH_HEAD."""
-    fetch_head = git.git_path('FETCH_HEAD')
-    if core.exists(fetch_head):
-        return git.fmt_merge_msg('--file', fetch_head)[STDOUT]
-    return "Merge branch '%s'" % revision
+def strip_remote(remotes, remote_branch):
+    for remote in remotes:
+        prefix = remote + '/'
+        if remote_branch.startswith(prefix):
+            return remote_branch[len(prefix):]
+    return remote_branch.split('/', 1)[-1]
