@@ -3,6 +3,8 @@ from __future__ import division, absolute_import, unicode_literals
 from PyQt4 import QtGui
 from PyQt4 import QtCore
 
+from cola import hotkeys
+
 
 def patch(obj, attr, value):
     if not hasattr(obj, attr):
@@ -10,9 +12,7 @@ def patch(obj, attr, value):
 
 
 def install():
-    set_contents_margins = (
-            lambda self, left, top, right, bottom: self.setMargin(left))
-
+    set_contents_margins = lambda self, *args: self.setMargin(max(args))
     patch(QtGui.QHBoxLayout, 'setContentsMargins', set_contents_margins)
     patch(QtGui.QVBoxLayout, 'setContentsMargins', set_contents_margins)
 
@@ -20,15 +20,15 @@ def install():
     patch(QtGui.QHBoxLayout, 'setMargin', set_margin)
     patch(QtGui.QVBoxLayout, 'setMargin', set_margin)
 
-    patch(QtGui.QKeySequence, 'Preferences', 'Ctrl+,')
+    patch(QtGui.QKeySequence, 'Preferences', hotkeys.PREFERENCES)
     patch(QtGui.QGraphicsItem, 'mapRectToScene', _map_rect_to_scene)
-    patch(QtCore.QCoreApplication, 'setStyleSheet', lambda *args: None)
 
 
 def add_search_path(prefix, path):
     if hasattr(QtCore.QDir, 'addSearchPath'):
         QtCore.QDir.addSearchPath(prefix, path)
 
+
 def set_common_dock_options(window):
     if not hasattr(window, 'setDockOptions'):
         return
