@@ -9,7 +9,7 @@ from PyQt4.QtCore import Qt
 from PyQt4.QtCore import SIGNAL
 
 from cola import gitcmds
-from cola.i18n import N_
+from cola import icons
 from cola import qtutils
 from cola import utils
 from cola.models import main
@@ -22,24 +22,44 @@ UPDATE_SIGNAL = 'update()'
 
 
 class CompletionLineEdit(text.HintedLineEdit):
+    """An lineedit with advanced completion abilities"""
 
-    def __init__(self, model, hint='', parent=None):
+    # Activation keys will cause a selected completion item to be chosen
+    ACTIVATION_KEYS = (Qt.Key_Return, Qt.Key_Enter)
+
+    # Navigation keys trigger signals that widgets can use for customization
+    NAVIGATION_KEYS = {
+            Qt.Key_Return: 'return()',
+            Qt.Key_Enter: 'enter()',
+            Qt.Key_Up: 'up()',
+            Qt.Key_Down: 'down()',
+    }
+
+    def __init__(self, model_factory, hint='', parent=None):
         text.HintedLineEdit.__init__(self, hint=hint, parent=parent)
+        # Tracks when the completion popup was active during key events
+        self._was_visible = False
+        # The most recently selected completion item
+        self._selection = None
 
-        self.setFont(qtutils.diff_font())
-
-        completion_model = model(self)
+        # Create a completion model
+        completion_model = model_factory(self)
         completer = Completer(completion_model, self)
         completer.setWidget(self)
         self._completer = completer
+        self._completion_model = completion_model
 
+        # The delegate highlights matching completion text in the popup widget
         self._delegate = HighlightDelegate(self)
-        self.connect(self, SIGNAL('textChanged(QString)'),
-                     self._text_changed)
-
         completer.popup().setItemDelegate(self._delegate)
+
+        self.connect(self, SIGNAL('textChanged(QString)'), self._text_changed)
+
         self.connect(self._completer, SIGNAL('activated(QString)'),
-                     self._complete)
+                     self.choose_completion)
+
+        self.connect(self._completion_model, SIGNAL('updated()'),
+                     self._completions_updated, Qt.QueuedConnection)
 
         self.connect(self, SIGNAL('destroyed(QObject*)'), self.dispose)
 
@@ -49,15 +69,26 @@ class CompletionLineEdit(text.HintedLineEdit):
     def dispose(self, *args):
         self._completer.dispose()
 
+    def was_visible(self):
+        """Was the popup visible during the last keypress event?"""
+        return self._was_visible
+
+    def completion_selection(self):
+        """Return the last completion's selection"""
+        return self._selection
+
+    def complete(self):
+        """Trigger the completion popup to appear and offer completions"""
+        self._completer.complete()
+
     def refresh(self):
+        """Refresh the completion model"""
         return self._completer.model().update()
 
     def popup(self):
+        """Return the completer's popup"""
         return self._completer.popup()
 
-    def value(self):
-        return ustr(self.text())
-
     def _is_case_sensitive(self, text):
         return bool([char for char in text if char.isupper()])
 
@@ -65,6 +96,7 @@ class CompletionLineEdit(text.HintedLineEdit):
         match_text = self._last_word()
         full_text = ustr(text)
         self._do_text_changed(full_text, match_text)
+        self.complete_last_word()
 
     def _do_text_changed(self, full_text, match_text):
         case_sensitive = self._is_case_sensitive(match_text)
@@ -78,9 +110,10 @@ class CompletionLineEdit(text.HintedLineEdit):
     def update_matches(self):
         text = self._last_word()
         case_sensitive = self._is_case_sensitive(text)
+        self._completer.setCompletionPrefix(text)
         self._completer.model().update_matches(case_sensitive)
 
-    def _complete(self, completion):
+    def choose_completion(self, completion):
         """
         This is the event handler for the QCompleter.activated(QString) signal,
         it is called when the user selects an item in the completer popup.
@@ -98,6 +131,7 @@ class CompletionLineEdit(text.HintedLineEdit):
         self.setText(text)
         self.emit(SIGNAL('changed()'))
         self._do_text_changed(text, '')
+        self.popup().hide()
 
     def _words(self):
         return utils.shell_split(self.value())
@@ -120,22 +154,9 @@ class CompletionLineEdit(text.HintedLineEdit):
             self.close_popup()
         return text.HintedLineEdit.event(self, event)
 
-    def do_completion(self):
-        self._completer.complete()
-        self.popup().setCurrentIndex(self._completer.model().index(0, 0))
-        self.popup().scrollToTop()
-
-    def keyReleaseEvent(self, event):
-
-        text.HintedLineEdit.keyReleaseEvent(self, event)
-
-        if event.key() in (Qt.Key_Tab, Qt.Key_Enter, Qt.Key_Return,
-                           Qt.Key_Escape):
-            return
-
-        prefix = self._last_word()
-        self._update_popup_items(prefix)
-        self.do_completion()
+    def complete_last_word(self):
+        self.update_matches()
+        self.complete()
 
     def close_popup(self):
         if self.popup().isVisible():
@@ -148,6 +169,79 @@ class CompletionLineEdit(text.HintedLineEdit):
         """
         self._completer.setCompletionPrefix(prefix)
 
+    def _completions_updated(self):
+        popup = self.popup()
+        if not popup.isVisible():
+            return
+        # Select the first item
+        idx = self._completion_model.index(0, 0)
+        selection = QtGui.QItemSelection(idx, idx)
+        mode = QtGui.QItemSelectionModel.Select
+        popup.selectionModel().select(selection, mode)
+
+    def selected_completion(self):
+        popup = self.popup()
+        if not popup.isVisible():
+            return None
+        model = popup.selectionModel()
+        indexes = model.selectedIndexes()
+        if not indexes:
+            return None
+        idx = indexes[0]
+        item = self._completion_model.itemFromIndex(idx)
+        if not item:
+            return
+        return ustr(item.text())
+
+    # Qt events
+    def keyPressEvent(self, event):
+        self._was_visible = visible = self.popup().isVisible()
+        key = event.key()
+        was_empty = not bool(self.value())
+
+        if visible:
+            self._selection = self.selected_completion()
+        else:
+            self._selection = None
+            if event.key() in self.ACTIVATION_KEYS:
+                event.accept()
+                return
+
+        result = text.HintedLineEdit.keyPressEvent(self, event)
+
+        # Backspace at the beginning of the line should hide the popup
+        if was_empty and visible and key == Qt.Key_Backspace:
+            self.popup().hide()
+        # Clearing a line should always emit a signal
+        is_empty = not bool(self.value())
+        if is_empty:
+            self.emit(SIGNAL('cleared()'))
+        return result
+
+    def keyReleaseEvent(self, event):
+        """React to release events, handle completion"""
+        key = event.key()
+        visible = self.was_visible()
+        if not visible:
+            # If it's a navigation key then emit a signal
+            try:
+                msg = self.NAVIGATION_KEYS[key]
+                event.accept()
+                self.emit(SIGNAL(msg))
+                return
+            except KeyError:
+                pass
+        # Run the real release event
+        result = text.HintedLineEdit.keyReleaseEvent(self, event)
+        # If the popup was visible and we have a selected popup item
+        # then choose that completion.
+        selection = self.completion_selection()
+        if visible and selection and key in self.ACTIVATION_KEYS:
+            self.choose_completion(selection)
+            self.emit(SIGNAL('activated()'))
+            return
+        return result
+
 
 class GatherCompletionsThread(QtCore.QThread):
 
@@ -166,7 +260,7 @@ class GatherCompletionsThread(QtCore.QThread):
             items = self.model.gather_matches(self.case_sensitive)
 
         if text is not None:
-            self.emit(SIGNAL('items_gathered'), items)
+            self.emit(SIGNAL('items_gathered(PyQt_PyObject)'), items)
 
 
 class HighlightDelegate(QtGui.QStyledItemDelegate):
@@ -235,6 +329,13 @@ class HighlightDelegate(QtGui.QStyledItemDelegate):
         painter.restore()
 
 
+def ref_sort_key(ref):
+    """Sort key function that causes shorter refs to sort first, but
+    alphabetizes refs of equal length (in order to make local branches sort
+    before remote ones)."""
+    return len(ref), ref
+
+
 class CompletionModel(QtGui.QStandardItemModel):
 
     def __init__(self, parent):
@@ -244,8 +345,9 @@ class CompletionModel(QtGui.QStandardItemModel):
         self.case_sensitive = False
 
         self.update_thread = GatherCompletionsThread(self)
-        self.connect(self.update_thread, SIGNAL('items_gathered'),
-                     self.apply_matches)
+        self.connect(self.update_thread,
+                     SIGNAL('items_gathered(PyQt_PyObject)'),
+                     self.apply_matches, Qt.QueuedConnection)
 
     def update(self):
         case_sensitive = self.update_thread.case_sensitive
@@ -266,12 +368,11 @@ class CompletionModel(QtGui.QStandardItemModel):
         return ((), (), set())
 
     def apply_matches(self, match_tuple):
-        self.match_tuple = match_tuple
         matched_refs, matched_paths, dirs = match_tuple
         QStandardItem = QtGui.QStandardItem
-        file_icon = qtutils.file_icon()
-        dir_icon = qtutils.dir_icon()
-        git_icon = qtutils.git_icon()
+
+        dir_icon = icons.directory()
+        git_icon = icons.cola()
 
         items = []
         for ref in matched_refs:
@@ -280,36 +381,37 @@ class CompletionModel(QtGui.QStandardItemModel):
             item.setIcon(git_icon)
             items.append(item)
 
+        from_filename = icons.from_filename
         for match in matched_paths:
             item = QStandardItem()
             item.setText(match)
             if match in dirs:
                 item.setIcon(dir_icon)
             else:
-                item.setIcon(file_icon)
+                item.setIcon(from_filename(match))
             items.append(item)
 
         self.clear()
         self.invisibleRootItem().appendRows(items)
+        self.emit(SIGNAL('updated()'))
 
 
-def filter_matches(match_text, candidates, case_sensitive):
+def filter_matches(match_text, candidates, case_sensitive,
+                   sort_key=lambda x: x):
     """Filter candidates and return the matches"""
 
     if case_sensitive:
-        transform = lambda x: x
-        keyfunc = lambda x: x.replace('.', '')
+        case_transform = lambda x: x
     else:
-        transform = lambda x: x.lower()
-        keyfunc = lambda x: x.replace('.', '').lower()
+        case_transform = lambda x: x.lower()
 
     if match_text:
-        matches = [r for r in candidates
-                    if transform(match_text) in transform(r)]
+        match_text = case_transform(match_text)
+        matches = [r for r in candidates if match_text in case_transform(r)]
     else:
         matches = list(candidates)
 
-    matches.sort(key=keyfunc)
+    matches.sort(key=lambda x: sort_key(case_transform(x)))
     return matches
 
 
@@ -332,7 +434,8 @@ class Completer(QtGui.QCompleter):
         self.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)
         self.setCaseSensitivity(Qt.CaseInsensitive)
 
-        self.connect(model, SIGNAL(UPDATE_SIGNAL), self.update)
+        self.connect(model, SIGNAL(UPDATE_SIGNAL),
+                     self.update, Qt.QueuedConnection)
         self.setModel(model)
 
     def update(self):
@@ -354,7 +457,8 @@ class GitCompletionModel(CompletionModel):
         model.add_observer(msg, self.emit_update)
 
     def gather_matches(self, case_sensitive):
-        refs = filter_matches(self.match_text, self.matches(), case_sensitive)
+        refs = filter_matches(self.match_text, self.matches(), case_sensitive,
+                              sort_key=ref_sort_key)
         return (refs, (), set())
 
     def emit_update(self):
@@ -381,6 +485,37 @@ class GitRefCompletionModel(GitCompletionModel):
         return model.local_branches + model.remote_branches + model.tags
 
 
+class GitPotentialBranchCompletionModel(GitCompletionModel):
+    """Completer for branches, tags, and potential branches"""
+
+    def __init__(self, parent):
+        GitCompletionModel.__init__(self, parent)
+
+    def matches(self):
+        model = self.main_model
+        remotes = model.remotes
+        remote_branches = model.remote_branches
+
+        ambiguous = set()
+        allnames = set(model.local_branches)
+        potential = []
+
+        for remote_branch in remote_branches:
+            branch = gitcmds.strip_remote(remotes, remote_branch)
+            if branch in allnames or branch == remote_branch:
+                ambiguous.add(branch)
+                continue
+            potential.append(branch)
+            allnames.add(branch)
+
+        potential_branches = [p for p in potential if p not in ambiguous]
+
+        return (model.local_branches +
+                potential_branches +
+                model.remote_branches +
+                model.tags)
+
+
 class GitBranchCompletionModel(GitCompletionModel):
     """Completer for remote branches"""
 
@@ -431,12 +566,13 @@ class GitStatusFilterCompletionModel(GitPathCompletionModel):
                 model.modified + model.untracked)
 
 
-class GitLogCompletionModel(GitRefCompletionModel):
-    """Completer for arguments suitable for git-log like commands"""
+class GitTrackedCompletionModel(GitPathCompletionModel):
+    """Completer for tracked files and folders"""
 
     def __init__(self, parent):
-        GitRefCompletionModel.__init__(self, parent)
-        self.connect(self, SIGNAL(UPDATE_SIGNAL), self.gather_paths)
+        GitPathCompletionModel.__init__(self, parent)
+        self.connect(self, SIGNAL(UPDATE_SIGNAL),
+                     self.gather_paths, Qt.QueuedConnection)
         self._paths = []
         self._updated = False
 
@@ -446,7 +582,31 @@ class GitLogCompletionModel(GitRefCompletionModel):
     def gather_matches(self, case_sensitive):
         if not self._paths:
             self.gather_paths()
-        refs = filter_matches(self.match_text, self.matches(), case_sensitive)
+
+        refs = []
+        paths, dirs = filter_path_matches(self.match_text, self._paths,
+                                          case_sensitive)
+        return (refs, paths, dirs)
+
+
+class GitLogCompletionModel(GitRefCompletionModel):
+    """Completer for arguments suitable for git-log like commands"""
+
+    def __init__(self, parent):
+        GitRefCompletionModel.__init__(self, parent)
+        self.connect(self, SIGNAL(UPDATE_SIGNAL),
+                     self.gather_paths, Qt.QueuedConnection)
+        self._paths = []
+        self._updated = False
+
+    def gather_paths(self):
+        self._paths = gitcmds.tracked_files()
+
+    def gather_matches(self, case_sensitive):
+        if not self._paths:
+            self.gather_paths()
+        refs = filter_matches(self.match_text, self.matches(), case_sensitive,
+                              sort_key=ref_sort_key)
         paths, dirs = filter_path_matches(self.match_text, self._paths,
                                           case_sensitive)
         has_doubledash = (self.match_text == '--' or
@@ -475,14 +635,16 @@ def bind_lineedit(model):
 # Concrete classes
 GitLogLineEdit = bind_lineedit(GitLogCompletionModel)
 GitRefLineEdit = bind_lineedit(GitRefCompletionModel)
+GitPotentialBranchLineEdit = bind_lineedit(GitPotentialBranchCompletionModel)
 GitBranchLineEdit = bind_lineedit(GitBranchCompletionModel)
 GitRemoteBranchLineEdit = bind_lineedit(GitRemoteBranchCompletionModel)
 GitStatusFilterLineEdit = bind_lineedit(GitStatusFilterCompletionModel)
+GitTrackedLineEdit = bind_lineedit(GitTrackedCompletionModel)
 
 
 class GitDialog(QtGui.QDialog):
 
-    def __init__(self, lineedit, title, button_text, parent):
+    def __init__(self, lineedit, title, button_text, parent, icon=None):
         QtGui.QDialog.__init__(self, parent)
         self.setWindowTitle(title)
         self.setMinimumWidth(333)
@@ -493,12 +655,10 @@ class GitDialog(QtGui.QDialog):
         self.lineedit = lineedit()
         self.setFocusProxy(self.lineedit)
 
-        self.ok_button = QtGui.QPushButton()
-        self.ok_button.setText(button_text)
-        self.ok_button.setIcon(qtutils.apply_icon())
-
-        self.close_button = QtGui.QPushButton()
-        self.close_button.setText(N_('Close'))
+        if icon is None:
+            icon = icons.ok()
+        self.ok_button = qtutils.create_button(text=button_text, icon=icon)
+        self.close_button = qtutils.close_button()
 
         self.button_layout = qtutils.hbox(defs.no_margin, defs.button_spacing,
                                           qtutils.STRETCH,
@@ -512,9 +672,9 @@ class GitDialog(QtGui.QDialog):
         qtutils.connect_button(self.ok_button, self.accept)
         qtutils.connect_button(self.close_button, self.reject)
 
-        self.connect(self.lineedit, SIGNAL('textChanged(const QString&)'),
+        self.connect(self.lineedit, SIGNAL('textChanged(QString)'),
                      self.text_changed)
-        self.connect(self.lineedit, SIGNAL('returnPressed()'), self.accept)
+        self.connect(self.lineedit, SIGNAL('return()'), self.accept)
 
         self.setWindowModality(Qt.WindowModal)
         self.ok_button.setEnabled(False)
@@ -529,8 +689,8 @@ class GitDialog(QtGui.QDialog):
         self.lineedit.setText(ref)
 
     @classmethod
-    def get(cls, title, button_text, parent, default=None):
-        dlg = cls(title, button_text, parent)
+    def get(cls, title, button_text, parent, default=None, icon=None):
+        dlg = cls(title, button_text, parent, icon=icon)
         if default:
             dlg.set_text(default)
 
@@ -556,20 +716,27 @@ class GitDialog(QtGui.QDialog):
 
 class GitRefDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent):
+    def __init__(self, title, button_text, parent, icon=None):
         GitDialog.__init__(self, GitRefLineEdit,
-                           title, button_text, parent)
+                           title, button_text, parent, icon=icon)
+
+
+class GitPotentialBranchDialog(GitDialog):
+
+    def __init__(self, title, button_text, parent, icon=None):
+        GitDialog.__init__(self, GitPotentialBranchLineEdit,
+                           title, button_text, parent, icon=icon)
 
 
 class GitBranchDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent):
+    def __init__(self, title, button_text, parent, icon=None):
         GitDialog.__init__(self, GitBranchLineEdit,
-                           title, button_text, parent)
+                           title, button_text, parent, icon=icon)
 
 
 class GitRemoteBranchDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent):
+    def __init__(self, title, button_text, parent, icon=None):
         GitDialog.__init__(self, GitRemoteBranchLineEdit,
-                           title, button_text, parent)
+                           title, button_text, parent, icon=icon)
